@@ -7,10 +7,13 @@ mod recover_error;
 #[cfg_attr(docsrs, doc(cfg(feature = "tls")))]
 mod tls;
 
-pub use conn::Connected;
+pub use conn::{Connected, TcpConnectInfo};
 #[cfg(feature = "tls")]
 pub use tls::ServerTlsConfig;
 
+#[cfg(feature = "tls")]
+pub use conn::TlsConnectInfo;
+
 #[cfg(feature = "tls")]
 use super::service::TlsAcceptor;
 
@@ -23,32 +26,37 @@ pub(crate) use tokio_rustls::server::TlsStream;
 use crate::transport::Error;
 
 use self::recover_error::RecoverError;
-use super::{
-    service::{GrpcTimeout, Or, Routes, ServerIo},
-    BoxFuture,
-};
-use crate::{body::BoxBody, request::ConnectionInfo};
+use super::service::{GrpcTimeout, Or, Routes, ServerIo};
+use crate::body::BoxBody;
+use bytes::Bytes;
 use futures_core::Stream;
 use futures_util::{
-    future::{self, Either as FutureEither, MapErr},
-    TryFutureExt,
+    future::{self, MapErr},
+    ready, TryFutureExt,
 };
-use http::{HeaderMap, Request, Response};
+use http::{Request, Response};
+use http_body::Body as _;
 use hyper::{server::accept, Body};
+use pin_project::pin_project;
 use std::{
     fmt,
     future::Future,
+    marker::PhantomData,
     net::SocketAddr,
+    pin::Pin,
     sync::Arc,
     task::{Context, Poll},
     time::Duration,
 };
 use tokio::io::{AsyncRead, AsyncWrite};
-use tower::{limit::concurrency::ConcurrencyLimitLayer, util::Either, Service, ServiceBuilder};
-use tracing_futures::{Instrument, Instrumented};
+use tower::{
+    layer::util::Identity, layer::Layer, limit::concurrency::ConcurrencyLimitLayer, util::Either,
+    Service, ServiceBuilder,
+};
 
-type BoxService = tower::util::BoxService<Request<Body>, Response<BoxBody>, crate::Error>;
-type TraceInterceptor = Arc<dyn Fn(&HeaderMap) -> tracing::Span + Send + Sync + 'static>;
+type BoxHttpBody = http_body::combinators::BoxBody<Bytes, crate::Error>;
+type BoxService = tower::util::BoxService<Request<Body>, Response<BoxHttpBody>, crate::Error>;
+type TraceInterceptor = Arc<dyn Fn(&http::Request<()>) -> tracing::Span + Send + Sync + 'static>;
 
 const DEFAULT_HTTP2_KEEPALIVE_TIMEOUT_SECS: u64 = 20;
 
@@ -61,7 +69,7 @@ const DEFAULT_HTTP2_KEEPALIVE_TIMEOUT_SECS: u64 = 20;
 /// reference implementation that should be a good starting point for anyone
 /// wanting to create a more complex and/or specific implementation.
 #[derive(Default, Clone)]
-pub struct Server {
+pub struct Server<L = Identity> {
     trace_interceptor: Option<TraceInterceptor>,
     concurrency_limit: Option<usize>,
     timeout: Option<Duration>,
@@ -75,12 +83,14 @@ pub struct Server {
     http2_keepalive_interval: Option<Duration>,
     http2_keepalive_timeout: Option<Duration>,
     max_frame_size: Option<u32>,
+    accept_http1: bool,
+    layer: L,
 }
 
 /// A stack based `Service` router.
 #[derive(Debug)]
-pub struct Router<A, B> {
-    server: Server,
+pub struct Router<A, B, L = Identity> {
+    server: Server<L>,
     routes: Routes<A, B, Request<Body>>,
 }
 
@@ -90,35 +100,29 @@ pub struct Router<A, B> {
 /// gRPC endpoints and can be consumed with the rest of the `tower`
 /// ecosystem.
 #[derive(Debug)]
-pub struct RouterService<A, B> {
-    router: Router<A, B>,
+pub struct RouterService<S> {
+    inner: S,
 }
 
-impl<A, B> Service<Request<Body>> for RouterService<A, B>
+impl<S> Service<Request<Body>> for RouterService<S>
 where
-    A: Service<Request<Body>, Response = Response<BoxBody>> + Clone + Send + 'static,
-    A::Future: Send + 'static,
-    A::Error: Into<crate::Error> + Send,
-    B: Service<Request<Body>, Response = Response<BoxBody>> + Clone + Send + 'static,
-    B::Future: Send + 'static,
-    B::Error: Into<crate::Error> + Send,
+    S: Service<Request<Body>, Response = Response<BoxBody>> + Clone + Send + 'static,
+    S::Future: Send + 'static,
+    S::Error: Into<crate::Error> + Send,
 {
     type Response = Response<BoxBody>;
     type Error = crate::Error;
 
     #[allow(clippy::type_complexity)]
-    type Future = FutureEither<
-        MapErr<A::Future, fn(A::Error) -> crate::Error>,
-        MapErr<B::Future, fn(B::Error) -> crate::Error>,
-    >;
+    type Future = MapErr<S::Future, fn(S::Error) -> crate::Error>;
 
+    #[inline]
     fn poll_ready(&mut self, _: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
         Poll::Ready(Ok(()))
     }
 
-    #[inline]
     fn call(&mut self, req: Request<Body>) -> Self::Future {
-        self.router.routes.call(req)
+        self.inner.call(req).map_err(Into::into)
     }
 }
 
@@ -140,12 +144,13 @@ impl Server {
     pub fn builder() -> Self {
         Server {
             tcp_nodelay: true,
+            accept_http1: false,
             ..Default::default()
         }
     }
 }
 
-impl Server {
+impl<L> Server<L> {
     /// Configure TLS for this server.
     #[cfg(feature = "tls")]
     #[cfg_attr(docsrs, doc(cfg(feature = "tls")))]
@@ -290,10 +295,25 @@ impl Server {
         }
     }
 
+    /// Allow this server to accept http1 requests.
+    ///
+    /// Accepting http1 requests is only useful when developing `grpc-web`
+    /// enabled services. If this setting is set to `true` but services are
+    /// not correctly configured to handle grpc-web requests, your server may
+    /// return confusing (but correct) protocol errors.
+    ///
+    /// Default is `false`.
+    pub fn accept_http1(self, accept_http1: bool) -> Self {
+        Server {
+            accept_http1,
+            ..self
+        }
+    }
+
     /// Intercept inbound headers and add a [`tracing::Span`] to each response future.
     pub fn trace_fn<F>(self, f: F) -> Self
     where
-        F: Fn(&HeaderMap) -> tracing::Span + Send + Sync + 'static,
+        F: Fn(&http::Request<()>) -> tracing::Span + Send + Sync + 'static,
     {
         Server {
             trace_interceptor: Some(Arc::new(f)),
@@ -305,7 +325,7 @@ impl Server {
     ///
     /// This will clone the `Server` builder and create a router that will
     /// route around different services.
-    pub fn add_service<S>(&mut self, svc: S) -> Router<S, Unimplemented>
+    pub fn add_service<S>(&mut self, svc: S) -> Router<S, Unimplemented, L>
     where
         S: Service<Request<Body>, Response = Response<BoxBody>>
             + NamedService
@@ -314,6 +334,7 @@ impl Server {
             + 'static,
         S::Future: Send + 'static,
         S::Error: Into<crate::Error> + Send,
+        L: Clone,
     {
         Router::new(self.clone(), svc)
     }
@@ -329,7 +350,7 @@ impl Server {
     pub fn add_optional_service<S>(
         &mut self,
         svc: Option<S>,
-    ) -> Router<Either<S, Unimplemented>, Unimplemented>
+    ) -> Router<Either<S, Unimplemented>, Unimplemented, L>
     where
         S: Service<Request<Body>, Response = Response<BoxBody>>
             + NamedService
@@ -338,6 +359,7 @@ impl Server {
             + 'static,
         S::Future: Send + 'static,
         S::Error: Into<crate::Error> + Send,
+        L: Clone,
     {
         let svc = match svc {
             Some(some) => Either::A(some),
@@ -346,33 +368,121 @@ impl Server {
         Router::new(self.clone(), svc)
     }
 
-    pub(crate) async fn serve_with_shutdown<S, I, F, IO, IE>(
+    /// Set the [Tower] [`Layer`] all services will be wrapped in.
+    ///
+    /// This enables using middleware from the [Tower ecosystem][eco].
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// # use tonic::transport::Server;
+    /// # use tower_service::Service;
+    /// use tower::timeout::TimeoutLayer;
+    /// use std::time::Duration;
+    ///
+    /// # let mut builder = Server::builder();
+    /// builder.layer(TimeoutLayer::new(Duration::from_secs(30)));
+    /// ```
+    ///
+    /// Note that timeouts should be set using [`Server::timeout`]. `TimeoutLayer` is only used
+    /// here as an example.
+    ///
+    /// You can build more complex layers using [`ServiceBuilder`]. Those layers can include
+    /// [interceptors]:
+    ///
+    /// ```
+    /// # use tonic::transport::Server;
+    /// # use tower_service::Service;
+    /// use tower::ServiceBuilder;
+    /// use std::time::Duration;
+    /// use tonic::{Request, Status, service::interceptor_fn};
+    ///
+    /// fn auth_interceptor(request: Request<()>) -> Result<Request<()>, Status> {
+    ///     if valid_credentials(&request) {
+    ///         Ok(request)
+    ///     } else {
+    ///         Err(Status::unauthenticated("invalid credentials"))
+    ///     }
+    /// }
+    ///
+    /// fn valid_credentials(request: &Request<()>) -> bool {
+    ///     // ...
+    ///     # true
+    /// }
+    ///
+    /// fn some_other_interceptor(request: Request<()>) -> Result<Request<()>, Status> {
+    ///     Ok(request)
+    /// }
+    ///
+    /// let layer = ServiceBuilder::new()
+    ///     .load_shed()
+    ///     .timeout(Duration::from_secs(30))
+    ///     .layer(interceptor_fn(auth_interceptor))
+    ///     .layer(interceptor_fn(some_other_interceptor))
+    ///     .into_inner();
+    ///
+    /// Server::builder().layer(layer);
+    /// ```
+    ///
+    /// [Tower]: https://github.com/tower-rs/tower
+    /// [`Layer`]: tower::layer::Layer
+    /// [eco]: https://github.com/tower-rs
+    /// [`ServiceBuilder`]: tower::ServiceBuilder
+    /// [interceptors]: crate::service::interceptor_fn
+    pub fn layer<NewLayer>(self, new_layer: NewLayer) -> Server<NewLayer> {
+        Server {
+            layer: new_layer,
+            trace_interceptor: self.trace_interceptor,
+            concurrency_limit: self.concurrency_limit,
+            timeout: self.timeout,
+            #[cfg(feature = "tls")]
+            tls: self.tls,
+            init_stream_window_size: self.init_stream_window_size,
+            init_connection_window_size: self.init_connection_window_size,
+            max_concurrent_streams: self.max_concurrent_streams,
+            tcp_keepalive: self.tcp_keepalive,
+            tcp_nodelay: self.tcp_nodelay,
+            http2_keepalive_interval: self.http2_keepalive_interval,
+            http2_keepalive_timeout: self.http2_keepalive_timeout,
+            max_frame_size: self.max_frame_size,
+            accept_http1: self.accept_http1,
+        }
+    }
+
+    pub(crate) async fn serve_with_shutdown<S, I, F, IO, IE, ResBody>(
         self,
         svc: S,
         incoming: I,
         signal: Option<F>,
     ) -> Result<(), super::Error>
     where
-        S: Service<Request<Body>, Response = Response<BoxBody>> + Clone + Send + 'static,
-        S::Future: Send + 'static,
-        S::Error: Into<crate::Error> + Send,
+        L: Layer<S>,
+        L::Service: Service<Request<Body>, Response = Response<ResBody>> + Clone + Send + 'static,
+        <<L as Layer<S>>::Service as Service<Request<Body>>>::Future: Send + 'static,
+        <<L as Layer<S>>::Service as Service<Request<Body>>>::Error: Into<crate::Error> + Send,
         I: Stream<Item = Result<IO, IE>>,
         IO: AsyncRead + AsyncWrite + Connected + Unpin + Send + 'static,
+        IO::ConnectInfo: Clone + Send + Sync + 'static,
         IE: Into<crate::Error>,
         F: Future<Output = ()>,
+        ResBody: http_body::Body<Data = Bytes> + Send + Sync + 'static,
+        ResBody::Error: Into<crate::Error>,
     {
-        let span = self.trace_interceptor.clone();
+        let trace_interceptor = self.trace_interceptor.clone();
         let concurrency_limit = self.concurrency_limit;
         let init_connection_window_size = self.init_connection_window_size;
         let init_stream_window_size = self.init_stream_window_size;
         let max_concurrent_streams = self.max_concurrent_streams;
         let timeout = self.timeout;
         let max_frame_size = self.max_frame_size;
+        let http2_only = !self.accept_http1;
 
         let http2_keepalive_interval = self.http2_keepalive_interval;
         let http2_keepalive_timeout = self
             .http2_keepalive_timeout
-            .unwrap_or(Duration::new(DEFAULT_HTTP2_KEEPALIVE_TIMEOUT_SECS, 0));
+            .unwrap_or_else(|| Duration::new(DEFAULT_HTTP2_KEEPALIVE_TIMEOUT_SECS, 0));
+
+        let svc = self.layer.layer(svc);
 
         let tcp = incoming::tcp_incoming(incoming, self);
         let incoming = accept::from_stream::<_, _, crate::Error>(tcp);
@@ -381,11 +491,12 @@ impl Server {
             inner: svc,
             concurrency_limit,
             timeout,
-            span,
+            trace_interceptor,
+            _io: PhantomData,
         };
 
         let server = hyper::Server::builder(incoming)
-            .http2_only(true)
+            .http2_only(http2_only)
             .http2_initial_connection_window_size(init_connection_window_size)
             .http2_initial_stream_window_size(init_stream_window_size)
             .http2_max_concurrent_streams(max_concurrent_streams)
@@ -407,8 +518,8 @@ impl Server {
     }
 }
 
-impl<S> Router<S, Unimplemented> {
-    pub(crate) fn new(server: Server, svc: S) -> Self
+impl<S, L> Router<S, Unimplemented, L> {
+    pub(crate) fn new(server: Server<L>, svc: S) -> Self
     where
         S: Service<Request<Body>, Response = Response<BoxBody>>
             + NamedService
@@ -432,7 +543,7 @@ impl<S> Router<S, Unimplemented> {
     }
 }
 
-impl<A, B> Router<A, B>
+impl<A, B, L> Router<A, B, L>
 where
     A: Service<Request<Body>, Response = Response<BoxBody>> + Clone + Send + 'static,
     A::Future: Send + 'static,
@@ -442,7 +553,7 @@ where
     B::Error: Into<crate::Error> + Send,
 {
     /// Add a new service to this router.
-    pub fn add_service<S>(self, svc: S) -> Router<S, Or<A, B, Request<Body>>>
+    pub fn add_service<S>(self, svc: S) -> Router<S, Or<A, B, Request<Body>>, L>
     where
         S: Service<Request<Body>, Response = Response<BoxBody>>
             + NamedService
@@ -471,10 +582,11 @@ where
     /// # Note
     /// Even when the argument given is `None` this will capture *all* requests to this service name.
     /// As a result, one cannot use this to toggle between two identically named implementations.
+    #[allow(clippy::type_complexity)]
     pub fn add_optional_service<S>(
         self,
         svc: Option<S>,
-    ) -> Router<Either<S, Unimplemented>, Or<A, B, Request<Body>>>
+    ) -> Router<Either<S, Unimplemented>, Or<A, B, Request<Body>>, L>
     where
         S: Service<Request<Body>, Response = Response<BoxBody>>
             + NamedService
@@ -503,27 +615,53 @@ where
     }
 
     /// Consume this [`Server`] creating a future that will execute the server
-    /// on [`tokio`]'s default executor.
+    /// on [tokio]'s default executor.
     ///
     /// [`Server`]: struct.Server.html
-    pub async fn serve(self, addr: SocketAddr) -> Result<(), super::Error> {
+    /// [tokio]: https://docs.rs/tokio
+    pub async fn serve<ResBody>(self, addr: SocketAddr) -> Result<(), super::Error>
+    where
+        L: Layer<Routes<A, B, Request<Body>>>,
+        L::Service: Service<Request<Body>, Response = Response<ResBody>> + Clone + Send + 'static,
+        <<L as Layer<Routes<A, B, Request<Body>>>>::Service as Service<Request<Body>>>::Future:
+            Send + 'static,
+        <<L as Layer<Routes<A, B, Request<Body>>>>::Service as Service<Request<Body>>>::Error:
+            Into<crate::Error> + Send,
+        ResBody: http_body::Body<Data = Bytes> + Send + Sync + 'static,
+        ResBody::Error: Into<crate::Error>,
+    {
         let incoming = TcpIncoming::new(addr, self.server.tcp_nodelay, self.server.tcp_keepalive)
             .map_err(super::Error::from_source)?;
         self.server
-            .serve_with_shutdown::<_, _, future::Ready<()>, _, _>(self.routes, incoming, None)
+            .serve_with_shutdown::<_, _, future::Ready<()>, _, _, ResBody>(
+                self.routes,
+                incoming,
+                None,
+            )
             .await
     }
 
     /// Consume this [`Server`] creating a future that will execute the server
-    /// on [`tokio`]'s default executor. And shutdown when the provided signal
+    /// on [tokio]'s default executor. And shutdown when the provided signal
     /// is received.
     ///
     /// [`Server`]: struct.Server.html
-    pub async fn serve_with_shutdown<F: Future<Output = ()>>(
+    /// [tokio]: https://docs.rs/tokio
+    pub async fn serve_with_shutdown<F: Future<Output = ()>, ResBody>(
         self,
         addr: SocketAddr,
         signal: F,
-    ) -> Result<(), super::Error> {
+    ) -> Result<(), super::Error>
+    where
+        L: Layer<Routes<A, B, Request<Body>>>,
+        L::Service: Service<Request<Body>, Response = Response<ResBody>> + Clone + Send + 'static,
+        <<L as Layer<Routes<A, B, Request<Body>>>>::Service as Service<Request<Body>>>::Future:
+            Send + 'static,
+        <<L as Layer<Routes<A, B, Request<Body>>>>::Service as Service<Request<Body>>>::Error:
+            Into<crate::Error> + Send,
+        ResBody: http_body::Body<Data = Bytes> + Send + Sync + 'static,
+        ResBody::Error: Into<crate::Error>,
+    {
         let incoming = TcpIncoming::new(addr, self.server.tcp_nodelay, self.server.tcp_keepalive)
             .map_err(super::Error::from_source)?;
         self.server
@@ -535,14 +673,30 @@ where
     /// the provided incoming stream of `AsyncRead + AsyncWrite`.
     ///
     /// [`Server`]: struct.Server.html
-    pub async fn serve_with_incoming<I, IO, IE>(self, incoming: I) -> Result<(), super::Error>
+    pub async fn serve_with_incoming<I, IO, IE, ResBody>(
+        self,
+        incoming: I,
+    ) -> Result<(), super::Error>
     where
         I: Stream<Item = Result<IO, IE>>,
         IO: AsyncRead + AsyncWrite + Connected + Unpin + Send + 'static,
+        IO::ConnectInfo: Clone + Send + Sync + 'static,
         IE: Into<crate::Error>,
+        L: Layer<Routes<A, B, Request<Body>>>,
+        L::Service: Service<Request<Body>, Response = Response<ResBody>> + Clone + Send + 'static,
+        <<L as Layer<Routes<A, B, Request<Body>>>>::Service as Service<Request<Body>>>::Future:
+            Send + 'static,
+        <<L as Layer<Routes<A, B, Request<Body>>>>::Service as Service<Request<Body>>>::Error:
+            Into<crate::Error> + Send,
+        ResBody: http_body::Body<Data = Bytes> + Send + Sync + 'static,
+        ResBody::Error: Into<crate::Error>,
     {
         self.server
-            .serve_with_shutdown::<_, _, future::Ready<()>, _, _>(self.routes, incoming, None)
+            .serve_with_shutdown::<_, _, future::Ready<()>, _, _, ResBody>(
+                self.routes,
+                incoming,
+                None,
+            )
             .await
     }
 
@@ -552,7 +706,7 @@ where
     /// gracefully shutdown the server.
     ///
     /// [`Server`]: struct.Server.html
-    pub async fn serve_with_incoming_shutdown<I, IO, IE, F>(
+    pub async fn serve_with_incoming_shutdown<I, IO, IE, F, ResBody>(
         self,
         incoming: I,
         signal: F,
@@ -560,8 +714,17 @@ where
     where
         I: Stream<Item = Result<IO, IE>>,
         IO: AsyncRead + AsyncWrite + Connected + Unpin + Send + 'static,
+        IO::ConnectInfo: Clone + Send + Sync + 'static,
         IE: Into<crate::Error>,
         F: Future<Output = ()>,
+        L: Layer<Routes<A, B, Request<Body>>>,
+        L::Service: Service<Request<Body>, Response = Response<ResBody>> + Clone + Send + 'static,
+        <<L as Layer<Routes<A, B, Request<Body>>>>::Service as Service<Request<Body>>>::Future:
+            Send + 'static,
+        <<L as Layer<Routes<A, B, Request<Body>>>>::Service as Service<Request<Body>>>::Error:
+            Into<crate::Error> + Send,
+        ResBody: http_body::Body<Data = Bytes> + Send + Sync + 'static,
+        ResBody::Error: Into<crate::Error>,
     {
         self.server
             .serve_with_shutdown(self.routes, incoming, Some(signal))
@@ -569,12 +732,23 @@ where
     }
 
     /// Create a tower service out of a router.
-    pub fn into_service(self) -> RouterService<A, B> {
-        RouterService { router: self }
+    pub fn into_service<ResBody>(self) -> RouterService<L::Service>
+    where
+        L: Layer<Routes<A, B, Request<Body>>>,
+        L::Service: Service<Request<Body>, Response = Response<ResBody>> + Clone + Send + 'static,
+        <<L as Layer<Routes<A, B, Request<Body>>>>::Service as Service<Request<Body>>>::Future:
+            Send + 'static,
+        <<L as Layer<Routes<A, B, Request<Body>>>>::Service as Service<Request<Body>>>::Error:
+            Into<crate::Error> + Send,
+        ResBody: http_body::Body<Data = Bytes> + Send + Sync + 'static,
+        ResBody::Error: Into<crate::Error>,
+    {
+        let inner = self.server.layer.layer(self.routes);
+        RouterService { inner }
     }
 }
 
-impl fmt::Debug for Server {
+impl<L> fmt::Debug for Server<L> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("Builder").finish()
     }
@@ -582,35 +756,69 @@ impl fmt::Debug for Server {
 
 struct Svc<S> {
     inner: S,
-    span: Option<TraceInterceptor>,
-    conn_info: ConnectionInfo,
+    trace_interceptor: Option<TraceInterceptor>,
 }
 
-impl<S> Service<Request<Body>> for Svc<S>
+impl<S, ResBody> Service<Request<Body>> for Svc<S>
 where
-    S: Service<Request<Body>, Response = Response<BoxBody>>,
+    S: Service<Request<Body>, Response = Response<ResBody>>,
     S::Error: Into<crate::Error>,
+    ResBody: http_body::Body<Data = Bytes> + Send + Sync + 'static,
+    ResBody::Error: Into<crate::Error>,
 {
-    type Response = Response<BoxBody>;
+    type Response = Response<BoxHttpBody>;
     type Error = crate::Error;
-
-    #[allow(clippy::type_complexity)]
-    type Future = MapErr<Instrumented<S::Future>, fn(S::Error) -> crate::Error>;
+    type Future = SvcFuture<S::Future>;
 
     fn poll_ready(&mut self, cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
         self.inner.poll_ready(cx).map_err(Into::into)
     }
 
     fn call(&mut self, mut req: Request<Body>) -> Self::Future {
-        let span = if let Some(trace_interceptor) = &self.span {
-            trace_interceptor(req.headers())
+        let span = if let Some(trace_interceptor) = &self.trace_interceptor {
+            let (parts, body) = req.into_parts();
+            let bodyless_request = Request::from_parts(parts, ());
+
+            let span = trace_interceptor(&bodyless_request);
+
+            let (parts, _) = bodyless_request.into_parts();
+            req = Request::from_parts(parts, body);
+
+            span
         } else {
             tracing::Span::none()
         };
 
-        req.extensions_mut().insert(self.conn_info.clone());
+        SvcFuture {
+            inner: self.inner.call(req),
+            span,
+        }
+    }
+}
 
-        self.inner.call(req).instrument(span).map_err(|e| e.into())
+#[pin_project]
+struct SvcFuture<F> {
+    #[pin]
+    inner: F,
+    span: tracing::Span,
+}
+
+impl<F, E, ResBody> Future for SvcFuture<F>
+where
+    F: Future<Output = Result<Response<ResBody>, E>>,
+    E: Into<crate::Error>,
+    ResBody: http_body::Body<Data = Bytes> + Send + Sync + 'static,
+    ResBody::Error: Into<crate::Error>,
+{
+    type Output = Result<Response<BoxHttpBody>, crate::Error>;
+
+    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
+        let this = self.project();
+        let _guard = this.span.enter();
+
+        let response: Response<ResBody> = ready!(this.inner.poll(cx)).map_err(Into::into)?;
+        let response = response.map(|body| body.map_err(Into::into).boxed());
+        Poll::Ready(Ok(response))
     }
 }
 
@@ -620,53 +828,76 @@ impl<S> fmt::Debug for Svc<S> {
     }
 }
 
-struct MakeSvc<S> {
+struct MakeSvc<S, IO> {
     concurrency_limit: Option<usize>,
     timeout: Option<Duration>,
     inner: S,
-    span: Option<TraceInterceptor>,
+    trace_interceptor: Option<TraceInterceptor>,
+    _io: PhantomData<fn() -> IO>,
 }
 
-impl<S> Service<&ServerIo> for MakeSvc<S>
+impl<S, ResBody, IO> Service<&ServerIo<IO>> for MakeSvc<S, IO>
 where
-    S: Service<Request<Body>, Response = Response<BoxBody>> + Clone + Send + 'static,
+    IO: Connected,
+    S: Service<Request<Body>, Response = Response<ResBody>> + Clone + Send + 'static,
     S::Future: Send + 'static,
     S::Error: Into<crate::Error> + Send,
+    ResBody: http_body::Body<Data = Bytes> + Send + Sync + 'static,
+    ResBody::Error: Into<crate::Error>,
 {
     type Response = BoxService;
     type Error = crate::Error;
-    type Future = BoxFuture<Self::Response, Self::Error>;
+    type Future = future::Ready<Result<Self::Response, Self::Error>>;
 
     fn poll_ready(&mut self, _cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
         Ok(()).into()
     }
 
-    fn call(&mut self, io: &ServerIo) -> Self::Future {
-        let conn_info = crate::request::ConnectionInfo {
-            remote_addr: io.remote_addr(),
-            peer_certs: io.peer_certs().map(Arc::new),
-        };
+    fn call(&mut self, io: &ServerIo<IO>) -> Self::Future {
+        let conn_info = io.connect_info();
 
         let svc = self.inner.clone();
         let concurrency_limit = self.concurrency_limit;
         let timeout = self.timeout;
-        let span = self.span.clone();
+        let trace_interceptor = self.trace_interceptor.clone();
 
-        Box::pin(async move {
-            let svc = ServiceBuilder::new()
-                .layer_fn(RecoverError::new)
-                .option_layer(concurrency_limit.map(ConcurrencyLimitLayer::new))
-                .layer_fn(|s| GrpcTimeout::new(s, timeout))
-                .service(svc);
+        let svc = ServiceBuilder::new()
+            .layer_fn(RecoverError::new)
+            .option_layer(concurrency_limit.map(ConcurrencyLimitLayer::new))
+            .layer_fn(|s| GrpcTimeout::new(s, timeout))
+            .service(svc);
 
-            let svc = BoxService::new(Svc {
+        let svc = ServiceBuilder::new()
+            .layer(BoxService::layer())
+            .map_request(move |mut request: Request<Body>| {
+                match &conn_info {
+                    tower::util::Either::A(inner) => {
+                        request.extensions_mut().insert(inner.clone());
+                    }
+                    tower::util::Either::B(inner) => {
+                        #[cfg(feature = "tls")]
+                        {
+                            request.extensions_mut().insert(inner.clone());
+                            request.extensions_mut().insert(inner.get_ref().clone());
+                        }
+
+                        #[cfg(not(feature = "tls"))]
+                        {
+                            // just a type check to make sure we didn't forget to
+                            // insert this into the extensions
+                            let _: &() = inner;
+                        }
+                    }
+                }
+
+                request
+            })
+            .service(Svc {
                 inner: svc,
-                span,
-                conn_info,
+                trace_interceptor,
             });
 
-            Ok(svc)
-        })
+        future::ready(Ok(svc))
     }
 }
 
@@ -691,7 +922,7 @@ impl Service<Request<Body>> for Unimplemented {
                 .status(200)
                 .header("grpc-status", "12")
                 .header("content-type", "application/grpc")
-                .body(BoxBody::empty())
+                .body(crate::body::empty_body())
                 .unwrap(),
         )
     }
